import enum


class ExtendedEnum(enum.Enum):

    @classmethod
    def list(cls):
        return list(map(lambda c: c.value, cls))


class IntegerDistributions(ExtendedEnum):
    uniform = 'uniform'


class ContinuousDistributions(ExtendedEnum):
    uniform = 'uniform'
    log_uniform = 'log-uniform'


class CategoricalDistributions(ExtendedEnum):
    choice = 'choice'




import numpy as np
from scipy import stats

from ..space.space_parameters import IntegerDistributions, ContinuousDistributions, CategoricalDistributions


class Integer(object):
    """class for hyperparameters search space of integer values"""

    def __init__(self, lower: int = None, upper: int = None, distribution: str = 'uniform'):
        """
        Parameters
        ----------
        lower: int
            lower bound of the possible values of the hyperparameter
        upper: int
            lower bound of the possible values of the hyperparameter
        distribution: "uniform"
            Distribution to sample initial population and mutation values
        """

        if not isinstance(lower, int):
            raise ValueError('lower bound must be an integer')

        if not isinstance(upper, int):
            raise ValueError('upper bound must be an integer')

        if lower > upper:
            raise ValueError('The upper bound can not be smaller that the lower bound')

        if distribution not in IntegerDistributions.list():
            raise ValueError(f"distribution must be one of {IntegerDistributions.list()}, got {distribution} instead")

        self.lower = lower
        self.upper = upper
        self.distribution = distribution

        if self.distribution == IntegerDistributions.uniform.value:
            self.rvs = stats.randint.rvs

    def sample(self):
        """Sample a random value from the assigned distribution"""

        return self.rvs(self.lower, self.upper + 1)


class Continuous(object):
    """class for hyperparameters search space of real values"""

    def __init__(self, lower: float = None, upper: float = None, distribution: str = 'uniform'):
        """
        Parameters
        ----------
        lower: int
            lower bound of the possible values of the hyperparameter
        upper: int
            lower bound of the possible values of the hyperparameter
        distribution: "uniform" or "log-uniform", default="uniform"
            Distribution to sample initial population and mutation values
        """

        if not isinstance(lower, (int, float)):
            raise ValueError('lower bound must be an integer or float')

        if not isinstance(upper, (int, float)):
            raise ValueError('upper bound must be an integer or float')

        if lower > upper:
            raise ValueError('The upper bound can not be smaller that the lower bound')

        if distribution not in ContinuousDistributions.list():
            raise ValueError(
                f"distribution must be one of {ContinuousDistributions.list()}, got {distribution} instead")

        self.lower = lower
        self.upper = upper
        self.distribution = distribution

        if self.distribution == ContinuousDistributions.uniform.value:
            self.rvs = stats.uniform.rvs
        elif self.distribution == ContinuousDistributions.log_uniform.value:
            self.rvs = stats.loguniform.rvs

    def sample(self):
        """Sample a random value from the assigned distribution"""

        return self.rvs(self.lower, self.upper)


class Categorical(object):
    """class for hyperparameters search space of categorical values"""

    def __init__(self, choices: list = None, priors: list = None, distribution: str = 'choice'):
        """
        Parameters
        ----------
        choices: list
            list with all the possible values of the hyperparameter
        priors: int
            list with the probability of sampling each element of the "choices", if not set gives equals probability
        distribution: "uniform" or "log-uniform", default="uniform"
            Distribution to sample initial population and mutation values
        """

        if not choices or not isinstance(choices, list):
            raise ValueError("choices must be a non empty list")

        if priors is None:
            self.priors = priors
        elif sum(priors) != 1:
            raise ValueError(f"The sum of the probabilities in the priors must be one, got {sum(priors)} instead")
        elif not len(priors) == len(choices):
            raise ValueError("priors and choices must have same size")
        else:
            self.priors = priors

        if distribution not in CategoricalDistributions.list():
            raise ValueError(
                f"distribution must be one of {CategoricalDistributions.list()}, got {distribution} instead")

        self.choices = choices
        self.distribution = distribution

        if self.distribution == CategoricalDistributions.choice.value:
            self.rvs = np.random.choice

    def sample(self):
        """Sample a random value from the assigned distribution"""

        return self.rvs(self.choices, p=self.priors)


def check_space(param_grid: dict = None):
    """

    Parameters
    ----------
    param_grid: dict
        Dictionary with the for {"hyperparameter_name": SpaceInstance}

    Returns
    -------
        Raises a Value Error if the dictionary does not have valid space instances
    """
    if not param_grid:
        raise ValueError(f"param_grid can not be empty")

    for key, value in param_grid.items():
        if not isinstance(value, (Integer, Categorical, Continuous)):
            raise ValueError(f"{key} must be a valid instance of Integer, Categorical or Continuous classes")


class Space(object):
    """Search space for all the models hyperparameters"""

    def __init__(self, param_grid: dict = None):
        check_space(param_grid)

        self.param_grid = param_grid

    @property
    def dimensions(self):
        return len(self.param_grid)

    @property
    def parameters(self):
        return list(self.param_grid.keys())

    def __len__(self):
        return self.dimensions

    def __getitem__(self, index):

        return self.param_grid[index]



import pytest

from ..space import Categorical, Integer, Continuous, Space


def test_sample_variables():
    my_categorical = Categorical(choices=['car', 'byc', 'house'], priors=[0.2, 0.1, 0.7])
    for _ in range(20):
        assert my_categorical.sample() in ['car', 'byc', 'house']

    my_continuous = Continuous(lower=0.01, upper=0.5, distribution='log-uniform')
    for _ in range(100):
        assert my_continuous.sample() <= 0.5
        assert my_continuous.sample() >= 0

    my_continuous = Continuous(lower=0.0, upper=0.5, distribution='uniform')
    for _ in range(100):
        assert my_continuous.sample() <= 0.5
        assert my_continuous.sample() >= 0

    my_integer = Integer(lower=5, upper=20, distribution='uniform')
    for _ in range(100):
        assert my_integer.sample() >= 5
        assert my_integer.sample() <= 20


def test_wrong_boundaries():
    with pytest.raises(Exception) as excinfo:
        my_continuous = Continuous(lower=10, upper=0.5)

    assert str(excinfo.value) == "The upper bound can not be smaller that the lower bound"

    with pytest.raises(Exception) as excinfo:
        my_integer = Integer(lower=10, upper=2)

    assert str(excinfo.value) == "The upper bound can not be smaller that the lower bound"


def test_wrong_distributions():
    with pytest.raises(Exception) as excinfo:
        my_continuous = Continuous(lower=2, upper=10, distribution='normal')
    assert str(excinfo.value) == "distribution must be one of ['uniform', 'log-uniform'], got normal instead"

    with pytest.raises(Exception) as excinfo:
        my_categorical = Categorical([True, False], distribution='sample')
    assert str(excinfo.value) == "distribution must be one of ['choice'], got sample instead"

    with pytest.raises(Exception) as excinfo:
        my_integer = Integer(lower=2, upper=10, distribution='log-uniform')
    assert str(excinfo.value) == "distribution must be one of ['uniform'], got log-uniform instead"


def test_categorical_bad_parameters():

    with pytest.raises(Exception) as excinfo:
        my_categorical = Categorical(priors=[0.1, 0.9])
    assert str(excinfo.value) == "choices must be a non empty list"

    with pytest.raises(Exception) as excinfo:
        my_categorical = Categorical(choices=[True, False], priors=[0.1, 0.8])
    assert str(excinfo.value) == "The sum of the probabilities in the priors must be one, got 0.9 instead"

    with pytest.raises(Exception) as excinfo:
        my_categorical = Categorical([True], priors=[0.1, 0.9])
    assert str(excinfo.value) == "priors and choices must have same size"


def test_check_space_fail():
    with pytest.raises(Exception) as excinfo:
        my_space = Space()
    assert str(excinfo.value) == "param_grid can not be empty"

    param_grid = {
        'min_weight_fraction_leaf': Continuous(lower=0.001, upper=0.5, distribution='log-uniform'),
        'max_leaf_nodes': Integer(lower=2, upper=35),
        'criterion': Categorical(choices=['gini', 'entropy']),
        'max_depth': range(10, 20)
    }

    with pytest.raises(Exception) as excinfo:
        my_space = Space(param_grid)
    assert str(excinfo.value) == "max_depth must be a valid instance of Integer, Categorical or Continuous classes"


def test_bad_data_types():

    with pytest.raises(Exception) as excinfo:
        Categorical((True, False))
    assert str(excinfo.value) == "choices must be a non empty list"

    with pytest.raises(Exception) as excinfo:
        Integer(5.4, 10)
    assert str(excinfo.value) == "lower bound must be an integer"

    with pytest.raises(Exception) as excinfo:
        Integer(5, 10.4)
    assert str(excinfo.value) == "upper bound must be an integer"

    with pytest.raises(Exception) as excinfo:
        Continuous([1], 10)
    assert str(excinfo.value) == "lower bound must be an integer or float"

    with pytest.raises(Exception) as excinfo:
        Continuous(5, [10.4])
    assert str(excinfo.value) == "upper bound must be an integer or float"



import warnings

import pytest
from sklearn.datasets import load_digits, load_boston
from sklearn.linear_model import SGDClassifier
from sklearn.tree import DecisionTreeClassifier
from sklearn.model_selection import train_test_split
from sklearn.utils.validation import check_is_fitted
from sklearn.tree import DecisionTreeRegressor
from sklearn.cluster import KMeans

from .. import GASearchCV
from ..space import Integer, Categorical, Continuous

data = load_digits()
label_names = data['target_names']
y = data['target']
X = data['data']

X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.33, random_state=42)


def test_expected_ga_results():
    clf = SGDClassifier(loss='log', fit_intercept=True)
    generations = 8
    evolved_estimator = GASearchCV(clf,
                                   cv=3,
                                   scoring='accuracy',
                                   population_size=6,
                                   generations=generations,
                                   tournament_size=3,
                                   elitism=False,
                                   keep_top_k=4,
                                   param_grid={'l1_ratio': Continuous(0, 1),
                                               'alpha': Continuous(1e-4, 1, distribution='log-uniform'),
                                               'average': Categorical([True, False]),
                                               'max_iter': Integer(700, 1000)},
                                   verbose=False,
                                   algorithm='eaSimple')

    evolved_estimator.fit(X_train, y_train)

    assert check_is_fitted(evolved_estimator) is None
    assert 'l1_ratio' in evolved_estimator.best_params
    assert 'alpha' in evolved_estimator.best_params
    assert 'average' in evolved_estimator.best_params
    assert len(evolved_estimator) == generations + 1  # +1 random initial population
    assert len(evolved_estimator.predict(X_test)) == len(X_test)
    assert evolved_estimator.score(X_train, y_train) >= 0
    assert len(evolved_estimator.decision_function(X_test)) == len(X_test)
    assert len(evolved_estimator.predict_proba(X_test)) == len(X_test)
    assert len(evolved_estimator.predict_log_proba(X_test)) == len(X_test)
    assert len(evolved_estimator.hof) == evolved_estimator.keep_top_k
    assert 'gen' in evolved_estimator[0]
    assert 'fitness_max' in evolved_estimator[0]
    assert 'fitness' in evolved_estimator[0]
    assert 'fitness_std' in evolved_estimator[0]
    assert 'fitness_min' in evolved_estimator[0]


def test_expected_ga_results_old_dicts():
    clf = SGDClassifier(loss='log', fit_intercept=True)
    generations = 8
    evolved_estimator = GASearchCV(clf,
                                   cv=3,
                                   scoring='accuracy',
                                   population_size=6,
                                   generations=generations,
                                   tournament_size=3,
                                   elitism=False,
                                   keep_top_k=4,
                                   continuous_parameters={'l1_ratio': (0, 1), 'alpha': (1e-4, 1)},
                                   categorical_parameters={'average': [True, False]},
                                   verbose=False,
                                   algorithm='eaSimple')

    evolved_estimator.fit(X_train, y_train)

    assert check_is_fitted(evolved_estimator) is None
    assert 'l1_ratio' in evolved_estimator.best_params
    assert 'alpha' in evolved_estimator.best_params
    assert 'average' in evolved_estimator.best_params
    assert len(evolved_estimator) == generations + 1  # +1 random initial population
    assert len(evolved_estimator.predict(X_test)) == len(X_test)
    assert evolved_estimator.score(X_train, y_train) >= 0
    assert len(evolved_estimator.decision_function(X_test)) == len(X_test)
    assert len(evolved_estimator.predict_proba(X_test)) == len(X_test)
    assert len(evolved_estimator.predict_log_proba(X_test)) == len(X_test)
    assert len(evolved_estimator.hof) == evolved_estimator.keep_top_k
    assert 'gen' in evolved_estimator[0]
    assert 'fitness_max' in evolved_estimator[0]
    assert 'fitness' in evolved_estimator[0]
    assert 'fitness_std' in evolved_estimator[0]
    assert 'fitness_min' in evolved_estimator[0]


def test_expected_ga_no_continuous():
    clf = DecisionTreeClassifier()
    generations = 8
    evolved_estimator = GASearchCV(clf,
                                   cv=3,
                                   scoring='accuracy',
                                   population_size=5,
                                   generations=generations,
                                   tournament_size=3,
                                   elitism=True,
                                   param_grid={'criterion': Categorical(['gini', 'entropy']),
                                               'max_depth': Integer(2, 20),
                                               'max_leaf_nodes': Integer(2, 30)},
                                   verbose=False)

    evolved_estimator.fit(X_train, y_train)

    assert check_is_fitted(evolved_estimator) is None
    assert 'criterion' in evolved_estimator.best_params
    assert 'max_depth' in evolved_estimator.best_params
    assert 'max_leaf_nodes' in evolved_estimator.best_params


def test_expected_ga_no_categorical():
    clf = DecisionTreeClassifier()
    generations = 8
    evolved_estimator = GASearchCV(clf,
                                   cv=3,
                                   scoring='accuracy',
                                   population_size=5,
                                   generations=generations,
                                   tournament_size=3,
                                   elitism=True,
                                   param_grid={'criterion': Categorical(['gini', 'entropy']),
                                               'max_depth': Integer(2, 20),
                                               'max_leaf_nodes': Integer(2, 30)},
                                   algorithm='eaMuCommaLambda',
                                   verbose=False)

    evolved_estimator.fit(X_train, y_train)

    assert check_is_fitted(evolved_estimator) is None
    assert 'criterion' in evolved_estimator.best_params
    assert 'max_depth' in evolved_estimator.best_params
    assert 'max_leaf_nodes' in evolved_estimator.best_params


def test_negative_criteria():
    data_boston = load_boston()

    y_boston = data_boston['target']
    X_boston = data_boston['data']

    X_train_b, X_test_b, y_train_b, y_test_b = train_test_split(X_boston, y_boston, test_size=0.33, random_state=42)

    clf = DecisionTreeRegressor()
    generations = 8
    evolved_estimator = GASearchCV(clf,
                                   cv=3,
                                   scoring='max_error',
                                   population_size=5,
                                   generations=generations,
                                   tournament_size=3,
                                   elitism=True,
                                   crossover_probability=0.9,
                                   mutation_probability=0.05,
                                   param_grid={'ccp_alpha': Continuous(0, 1),
                                               'criterion': Categorical(['mse', 'mae']),
                                               'max_depth': Integer(2, 20),
                                               'min_samples_split': Integer(2, 30)},
                                   criteria='min',
                                   n_jobs=-1)

    evolved_estimator.fit(X_train_b, y_train_b)

    assert check_is_fitted(evolved_estimator) is None
    assert 'ccp_alpha' in evolved_estimator.best_params
    assert 'criterion' in evolved_estimator.best_params
    assert 'max_depth' in evolved_estimator.best_params
    assert 'min_samples_split' in evolved_estimator.best_params
    assert len(evolved_estimator.predict(X_test_b)) == len(X_test_b)
    assert evolved_estimator.score(X_train_b, y_train_b) >= 0


def test_wrong_criteria():
    clf = SGDClassifier(loss='log', fit_intercept=True)
    generations = 8
    with pytest.raises(Exception) as excinfo:
        evolved_estimator = GASearchCV(clf,
                                       cv=3,
                                       scoring='accuracy',
                                       population_size=5,
                                       generations=generations,
                                       tournament_size=3,
                                       elitism=False,
                                       param_grid={'l1_ratio': Continuous(0, 1),
                                                   'alpha': Continuous(1e-4, 1),
                                                   'average': Categorical([True, False])},
                                       verbose=False,
                                       criteria='maximization')
    assert str(excinfo.value) == "Criteria must be one of ['max', 'min'], got maximization instead"


def test_wrong_estimator():
    clf = KMeans()
    generations = 8
    with pytest.raises(Exception) as excinfo:
        evolved_estimator = GASearchCV(clf,
                                       cv=3,
                                       scoring='accuracy',
                                       population_size=5,
                                       generations=generations,
                                       tournament_size=3,
                                       elitism=False,
                                       param_grid={'l1_ratio': Continuous(0, 1),
                                                   'alpha': Continuous(1e-4, 1),
                                                   'average': Categorical([True, False])},
                                       verbose=False,
                                       criteria='maximization')
    assert str(excinfo.value) == "KMeans() is not a valid Sklearn classifier or regressor"


def test_wrong_get_item():
    clf = SGDClassifier(loss='log', fit_intercept=True)
    generations = 8
    evolved_estimator = GASearchCV(clf,
                                   cv=3,
                                   scoring='accuracy',
                                   population_size=5,
                                   generations=generations,
                                   tournament_size=3,
                                   elitism=False,
                                   param_grid={'l1_ratio': Continuous(0, 1),
                                               'alpha': Continuous(1e-4, 1),
                                               'average': Categorical([True, False])},
                                   verbose=False,
                                   criteria='max')
    with pytest.raises(Exception) as excinfo:
        value = evolved_estimator[0]
    assert str(
        excinfo.value) == "This GASearchCV instance is not fitted yet. Call 'fit' with appropriate arguments before " \
                          "using this estimator."


def test_iterator():
    clf = DecisionTreeClassifier()
    generations = 8
    evolved_estimator = GASearchCV(clf,
                                   cv=3,
                                   scoring='accuracy',
                                   population_size=5,
                                   generations=generations,
                                   tournament_size=3,
                                   elitism=True,
                                   param_grid={'min_weight_fraction_leaf': Continuous(0, 0.5),
                                               'max_depth': Integer(2, 20),
                                               'max_leaf_nodes': Integer(2, 30)},
                                   verbose=False)
    evolved_estimator.fit(X_train, y_train)

    i = iter(evolved_estimator)
    assert next(i) == evolved_estimator[0]
    assert next(i) == evolved_estimator[1]


def test_wrong_algorithm():
    clf = SGDClassifier(loss='log', fit_intercept=True)
    generations = 8
    evolved_estimator = GASearchCV(clf,
                                   cv=3,
                                   scoring='accuracy',
                                   population_size=5,
                                   generations=generations,
                                   tournament_size=3,
                                   elitism=False,
                                   param_grid={'l1_ratio': Continuous(0, 1),
                                               'alpha': Continuous(1e-4, 1),
                                               'average': Categorical([True, False])},
                                   verbose=False,
                                   criteria='max',
                                   algorithm='genetic')
    with pytest.raises(Exception) as excinfo:
        evolved_estimator.fit(X_train, y_train)
    assert str(
        excinfo.value) == "The algorithm genetic is not supported, please select one from ['eaSimple', 'eaMuPlusLambda', 'eaMuCommaLambda']"


def test_ga_old_dicts_warnings():
    clf = SGDClassifier(loss='log', fit_intercept=True)
    generations = 8

    with pytest.warns(DeprecationWarning) as record:
        evolved_estimator = GASearchCV(clf,
                                       cv=3,
                                       scoring='accuracy',
                                       population_size=5,
                                       generations=generations,
                                       tournament_size=3,
                                       elitism=False,
                                       keep_top_k=4,
                                       continuous_parameters={'l1_ratio': (0, 1), 'alpha': (1e-4, 1)},
                                       categorical_parameters={'average': [True, False]},
                                       integer_parameters={'max_iter': (700, 1000)},
                                       param_grid={'l1_ratio': Continuous(0, 1),
                                                   'alpha': Continuous(1e-4, 1, distribution='log-uniform'),
                                                   'average': Categorical([True, False]),
                                                   'max_iter': Integer(700, 1000)},
                                       verbose=False,
                                       algorithm='eaSimple')

    # check that 4 warning was raised
    assert len(record) == 4
    # check that the message matches
    assert record[0].message.args[0] == "Warning, 'continuous_parameters' is deprecated and will be removed from " \
                                        "package in version 0.4.0, you should use 'param_grid' instead"

    assert record[1].message.args[0] == "Warning, 'categorical_parameters' is deprecated and will be removed from " \
                                        "package in version 0.4.0, you should use 'param_grid' instead"

    assert record[2].message.args[0] == "Warning, 'integer_parameters' is deprecated and will be removed from package " \
                                        "in version 0.4.0, you should use 'param_grid' instead"

    assert record[3].message.args[0] == "Warning, found parameters both in param_grid and older dictionary " \
                                        "parameters, algorithm is only going to use param_grid"


def test_no_param_grid():
    clf = SGDClassifier(loss='log', fit_intercept=True)
    generations = 8
    with pytest.raises(Exception) as excinfo:
        evolved_estimator = GASearchCV(clf,
                                       cv=3,
                                       scoring='accuracy',
                                       population_size=12,
                                       generations=generations,
                                       tournament_size=3,
                                       elitism=False,
                                       verbose=False,
                                       criteria='max')

    assert str(excinfo.value) == "param_grid can not be empty"



import warnings

import numpy as np
import random
from deap import base, creator, tools, algorithms
from sklearn.base import clone, ClassifierMixin, RegressorMixin
from sklearn.model_selection import cross_val_score
from sklearn.base import is_classifier, is_regressor
from sklearn.utils.metaestimators import if_delegate_has_method
from sklearn.utils.validation import check_array, check_is_fitted
from sklearn.metrics import check_scoring
from sklearn.exceptions import NotFittedError

from .parameters import Algorithms, Criteria
from .space import Space, Continuous, Integer, Categorical


def create_space(continuous_parameters, categorical_parameters, integer_parameters, param_grid):
    local_space = {}

    if continuous_parameters:
        warnings.warn("Warning, 'continuous_parameters' is deprecated and will be removed from package in version "
                      "0.4.0, you should use 'param_grid' instead", DeprecationWarning)
        for key, value in continuous_parameters.items():
            local_space[key] = Continuous(lower=value[0], upper=value[1])

    if categorical_parameters:
        warnings.warn("Warning, 'categorical_parameters' is deprecated and will be removed from package in version "
                      "0.4.0, you should use 'param_grid' instead", DeprecationWarning)
        for key, value in categorical_parameters.items():
            local_space[key] = Categorical(choices=value)

    if integer_parameters:
        warnings.warn("Warning, 'integer_parameters' is deprecated and will be removed from package in version 0.4.0, "
                      "you should use 'param_grid' instead", DeprecationWarning)
        for key, value in integer_parameters.items():
            local_space[key] = Integer(lower=value[0], upper=value[1])

    if bool(local_space) and bool(param_grid):
        warnings.warn("Warning, found parameters both in param_grid and older dictionary parameters, "
                      "algorithm is only going to use param_grid")
        space = Space(param_grid)
    elif bool(param_grid):
        space = Space(param_grid)
    elif bool(local_space):
        space = Space(local_space)
    else:
        space = Space()

    return space


class GASearchCV(ClassifierMixin, RegressorMixin):
    """Scikit-learn Hyperparameters tuning using evolutionary algorithms."""

    def __init__(self,
                 estimator,
                 cv: int = 3,
                 scoring=None,
                 population_size: int = 20,
                 generations: int = 40,
                 crossover_probability: float = 0.8,
                 mutation_probability: float = 0.1,
                 tournament_size: int = 3,
                 elitism: bool = True,
                 verbose: bool = True,
                 keep_top_k: int = 1,
                 param_grid: dict = None,
                 continuous_parameters: dict = None,
                 categorical_parameters: dict = None,
                 integer_parameters: dict = None,
                 criteria: str = 'max',
                 algorithm: str = 'eaMuPlusLambda',
                 n_jobs: int = 1):
        """

        Parameters
        ----------
        estimator: Sklearn Classifier or Regressor
        cv: int, number of splits used for calculating cross_val_score
        scoring: string, Scoring function to use as fitness value
        population_size: int, size of the population
        generations: int, number of generations to run the genetic algorithm
        crossover_probability: float, probability of crossover operation
        mutation_probability: float, probability of child mutation
        tournament_size: number of chromosomes to perform tournament selection
        elitism: bool, if true takes the |tournament_size| best solution to the next generation
        verbose: bool, if true, shows the metrics on the optimization routine
        keep_top_k: int, number of best solutions to keep in the hof object
        param_grid: dict, grid with the parameters to tune, expects as values of each key a sklearn_genetic.space Integer, Categorical or Continuous
        continuous_parameters: dict, continuous parameters to tune, expected a list or tuple with the range (min,max) to search
        categorical_parameters: dict, categorical parameters to tune, expected a list with the possible options to choose
        integer_parameters: dict, integers parameters to tune, expected a list or tuple with the range (min,max) to search
        criteria: str, 'max' if a higher scoring metric is better, 'min' otherwise
        algorithm: str, accepts 'eaSimple', 'eaMuPlusLambda' or 'eaMuCommaLambda' as optimization routines. See more details in the deap algorithms documentation
        n_jobs: int, Number of jobs to run in parallel during the cross validation scoring
        """

        self.estimator = clone(estimator)
        self.toolbox = base.Toolbox()
        self.cv = cv
        self.scoring = scoring
        self.pop_size = population_size
        self.generations = generations
        self.crossover_probability = crossover_probability
        self.mutation_probability = mutation_probability
        self.tournament_size = tournament_size
        self.elitism = elitism
        self.verbose = verbose
        self.keep_top_k = keep_top_k
        self.param_grid = param_grid
        self.algorithm = algorithm
        self.n_jobs = n_jobs
        self.creator = creator
        self.logbook = None
        self.history = None
        self.X = None
        self.Y = None
        self.best_params = None
        self.hof = None
        self.X_predict = None

        if not is_classifier(self.estimator) and not is_regressor(self.estimator):
            raise ValueError("{} is not a valid Sklearn classifier or regressor".format(self.estimator))

        if criteria not in Criteria.list():
            raise ValueError(f"Criteria must be one of {Criteria.list()}, got {criteria} instead")
        elif criteria == Criteria.max.value:
            self.criteria_sign = 1
        elif criteria == Criteria.min.value:
            self.criteria_sign = -1

        self.space = create_space(continuous_parameters=continuous_parameters,
                                  categorical_parameters=categorical_parameters,
                                  integer_parameters=integer_parameters,
                                  param_grid=param_grid)

    def register(self):

        self.creator.create("FitnessMax", base.Fitness, weights=[1.0])
        self.creator.create("Individual", list, fitness=creator.FitnessMax)

        attributes = []

        for parameter, dimension in self.space.param_grid.items():
            self.toolbox.register(f"{parameter}", dimension.sample)
            attributes.append(getattr(self.toolbox, parameter))

        IND_SIZE = 1

        self.toolbox.register("individual",
                              tools.initCycle, creator.Individual,
                              tuple(attributes), n=IND_SIZE)

        self.toolbox.register("population", tools.initRepeat, list, self.toolbox.individual)

        self.toolbox.register("mate", tools.cxTwoPoint)
        self.toolbox.register("mutate", self.mutate)
        if self.elitism:
            self.toolbox.register("select", tools.selTournament, tournsize=self.tournament_size)
        else:
            self.toolbox.register("select", tools.selRoulette)

        self.toolbox.register("evaluate", self.evaluate)

    def mutate(self, individual):

        gen = random.randrange(0, len(self.space))
        parameter_idx = self.space.parameters[gen]
        parameter = self.space[parameter_idx]

        individual[gen] = parameter.sample()

        return [individual]

    def evaluate(self, individual):
        current_generation_params = {key: individual[n] for n, key in enumerate(self.space.parameters)}

        self.estimator.set_params(**current_generation_params)
        cv_scores = cross_val_score(self.estimator,
                                    self.X_, self.Y_,
                                    cv=self.cv,
                                    scoring=self.scoring,
                                    n_jobs=self.n_jobs)
        score = np.mean(cv_scores)

        current_generation_params['score'] = score

        self.logbook.record(parameters=current_generation_params)

        return [self.criteria_sign * score]

    @if_delegate_has_method(delegate='estimator')
    def fit(self, X, y):
        """
        Main method of GASearchCV, optimize the hyper parameters of the given estimator
        Parameters
        ----------
        X: training samples to learn from
        y: training labels for each X obversation

        Returns

        fitted sklearn Regressor or Classifier
        -------

        """
        scorer = check_scoring(self.estimator, scoring=self.scoring)

        self.X_ = X
        self.Y_ = y

        self.register()

        pop = self.toolbox.population(n=self.pop_size)
        hof = tools.HallOfFame(self.keep_top_k)

        stats = tools.Statistics(lambda ind: ind.fitness.values)
        stats.register("fitness", np.mean)
        stats.register("fitness_std", np.std)
        stats.register("fitness_max", np.max)
        stats.register("fitness_min", np.min)

        self.logbook = tools.Logbook()

        pop, log = self._select_algorithm(pop=pop, stats=stats, hof=hof)

        self.best_params = {key: hof[0][n] for n, key in enumerate(self.space.parameters)}

        self.hof = {k: {key: hof[k][n] for n, key in enumerate(self.space.parameters)} for k in range(self.keep_top_k)}

        self.history = {"gen": log.select("gen"),
                        "fitness": log.select("fitness"),
                        "fitness_std": log.select("fitness_std"),
                        "fitness_max": log.select("fitness_max"),
                        "fitness_min": log.select("fitness_min")}

        self.estimator.set_params(**self.best_params)
        self.estimator.fit(self.X_, self.Y_)

        del self.creator.FitnessMax
        del self.creator.Individual

        return self

    def _select_algorithm(self, pop, stats, hof):

        if self.algorithm == Algorithms.eaSimple.value:

            pop, log = algorithms.eaSimple(pop, self.toolbox,
                                           cxpb=self.crossover_probability,
                                           stats=stats,
                                           mutpb=self.mutation_probability,
                                           ngen=self.generations,
                                           halloffame=hof,
                                           verbose=self.verbose)

        elif self.algorithm == Algorithms.eaMuPlusLambda.value:

            pop, log = algorithms.eaMuPlusLambda(pop, self.toolbox,
                                                 mu=self.pop_size,
                                                 lambda_=2 * self.pop_size,
                                                 cxpb=self.crossover_probability,
                                                 stats=stats,
                                                 mutpb=self.mutation_probability,
                                                 ngen=self.generations,
                                                 halloffame=hof,
                                                 verbose=self.verbose)

        elif self.algorithm == Algorithms.eaMuCommaLambda.value:
            pop, log = algorithms.eaMuCommaLambda(pop, self.toolbox,
                                                  mu=self.pop_size,
                                                  lambda_=2 * self.pop_size,
                                                  cxpb=self.crossover_probability,
                                                  stats=stats,
                                                  mutpb=self.mutation_probability,
                                                  ngen=self.generations,
                                                  halloffame=hof,
                                                  verbose=self.verbose)

        else:
            raise ValueError(
                f"The algorithm {self.algorithm} is not supported, please select one from {Algorithms.list()}")

        return pop, log

    @property
    def fitted(self):
        try:
            check_is_fitted(self.estimator)
            is_fitted = True
        except Exception as e:
            is_fitted = False

        has_history = bool(self.history)
        return all([is_fitted, has_history])

    def __getitem__(self, index):
        """

        Parameters
        ----------
        index: slice required to get

        Returns
        -------
        Best solution of the iteration corresponding to the index number
        """
        if not self.fitted:
            raise NotFittedError(
                f"This GASearchCV instance is not fitted yet. Call 'fit' with appropriate arguments before using this estimator.")

        return {"gen": self.history['gen'][index],
                "fitness": self.history['fitness'][index],
                "fitness_std": self.history['fitness_std'][index],
                "fitness_max": self.history['fitness_max'][index],
                "fitness_min": self.history['fitness_min'][index]}

    def __iter__(self):
        self.n = 0
        return self

    def __next__(self):
        """
        Returns
        -------
        Iteration over the statistics found in each generation
        """
        if self.n < self.generations + 1:
            result = self.__getitem__(self.n)
            self.n += 1
            return result
        else:
            raise StopIteration

    def __len__(self):
        """
        Returns
        -------
        Number of generations fitted
        """
        return self.generations + 1

    @if_delegate_has_method(delegate='estimator')
    def predict(self, X):
        X = check_array(X)
        return self.estimator.predict(X)

    @if_delegate_has_method(delegate='estimator')
    def score(self, X, y):
        X = check_array(X)
        return self.estimator.score(X, y)

    @if_delegate_has_method(delegate='estimator')
    def decision_function(self, X):
        X = check_array(X)
        return self.estimator.decision_function(X)

    @if_delegate_has_method(delegate='estimator')
    def predict_proba(self, X):
        X = check_array(X)
        return self.estimator.predict_proba(X)

    @if_delegate_has_method(delegate='estimator')
    def predict_log_proba(self, X):
        X = check_array(X)
        return self.estimator.predict_log_proba(X)



import seaborn as sns

sns.set_style("whitegrid")


def plot_fitness_evolution(estimator):
    """
    Parameters
    ----------
    estimator: Fitted GASearchCV

    Returns
    plot with the fitness value in each generation
    -------

    """
    fitness_history = estimator.history["fitness"]

    palette = sns.color_palette("rocket")
    sns.set(rc={'figure.figsize': (12, 15)})

    ax = sns.lineplot(x=range(len(estimator)), y=fitness_history,
                      markers=True,
                      palette=palette)
    ax.set_title('Fitness average evolution over generations')

    ax.set(xlabel='generations', ylabel=f'fitness ({estimator.scoring})')
    return ax



from sklearn_genetic.utils import plot_fitness_evolution
from sklearn_genetic import GASearchCV
from sklearn.datasets import load_boston
from sklearn.model_selection import train_test_split
from sklearn.tree import DecisionTreeRegressor

data = load_boston()

y = data['target']
X = data['data']

X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.33, random_state=42)

clf = DecisionTreeRegressor()

evolved_estimator = GASearchCV(clf,
                               cv=3,
                               scoring='r2',
                               population_size=8,
                               generations=60,
                               tournament_size=3,
                               elitism=True,
                               crossover_probability=0.9,
                               mutation_probability=0.05,
                               continuous_parameters={'ccp_alpha': (0, 1)},
                               categorical_parameters={'criterion': ['mse', 'mae']},
                               integer_parameters={'max_depth': (2, 20), 'min_samples_split': (2, 30)},
                               criteria='max',
                               n_jobs=-1)

evolved_estimator.fit(X_train, y_train)


def test_plot_created():
    plot = plot_fitness_evolution(evolved_estimator)




import enum


class ExtendedEnum(enum.Enum):

    @classmethod
    def list(cls):
        return list(map(lambda c: c.value, cls))


class Algorithms(ExtendedEnum):
    eaSimple = 'eaSimple'
    eaMuPlusLambda = 'eaMuPlusLambda'
    eaMuCommaLambda = 'eaMuCommaLambda'


class Criteria(ExtendedEnum):
    max = 'max'
    min = 'min'



import pathlib
from setuptools import setup, find_packages

# python setup.py sdist bdist_wheel
# twine upload --skip-existing --repository-url https://test.pypi.org/legacy/ dist/*

HERE = pathlib.Path(__file__).parent

README = (HERE / "README.md").read_text()
setup(
    name="sklearn-genetic-opt",
    version="0.3.0dev1",
    description="Scikit-lean models hyperparameters tuning, using evolutionary algorithms",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/rodrigo-arenas/Sklearn-genetic-opt",
    author="Rodrigo Arenas",
    author_email="rodrigo.arenas456@gmail.com",
    license="MIT",
    classifiers=[
        'License :: OSI Approved :: MIT License',
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",

    ],
    packages=find_packages(include=['sklearn_genetic', 'sklearn_genetic.*']),
    install_requires=['scikit-learn>=0.21.3', 'numpy>=1.13.3', 'seaborn>=0.11.1', 'deap>=1.3.1', 'pydantic>=1.8.2'],
    python_requires=">=3.6",
    include_package_data=True,
)
